/* vim: set expandtab ts=4 sw=4 : */
/*
    subidfs.c
    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
    Copyright (C) 2006-2007  Phil Snowberger <psnowber@cse.nd.edu>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.

*/

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#endif

#define _FILE_OFFSET_BITS 64 
#define FUSE_USE_VERSION 25

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include "subid.h"
#include "util_uid.h"
#include <stdlib.h>
#include <sys/inotify.h>
#include <signal.h>
#include <pthread.h>
#include <sys/poll.h>

#define USE_INOTIFY

/*#define DEBUG*/

void subidfs_read_data();
struct child_parent* etc_subusers_contents;

#ifdef USE_INOTIFY
#define EVENT_SIZE (sizeof (struct inotify_event))
#define INOTIFY_BUF_LEN (1024 * (EVENT_SIZE + 16))
struct pollfd inotify_fd;
char inotify_buf[INOTIFY_BUF_LEN];
int using_inotify;
int wd;
#endif /* USE_INOTIFY */
struct stat st;
time_t last_mtime = 0;

pthread_mutex_t esc_mutex = PTHREAD_MUTEX_INITIALIZER;

static void update_etc_subusers_contents() {
    int ret;
    ssize_t len;
    int i;

/*fprintf(stderr, "pid %d tid %d: acquiring esc_mutex\n", getpid(), (int)pthread_self());*/
    pthread_mutex_lock(&esc_mutex);
/*fprintf(stderr, "pid %d tid %d: acquired esc_mutex\n", getpid(), (int)pthread_self());*/
#ifdef USE_INOTIFY
    if (using_inotify) {
        ret = poll(&inotify_fd, 1, 0);
        if (ret > 0) {
            len = read(inotify_fd.fd, inotify_buf, INOTIFY_BUF_LEN);
            if (len > 0) {
                i = 0;
                while (i < len) {
                    struct inotify_event* event;
                    event = (struct inotify_event*) &inotify_buf[i];
                    if (wd == event->wd) {
/*fprintf(stderr, "pid %d tid %d: inotify says we should update\n", getpid(), (int)pthread_self());*/

                        subidfs_read_data();
/*fprintf(stderr, "pid %d tid %d: finished updating from inotify\n", getpid(), (int)pthread_self());*/
                    }
                    i += (sizeof (struct inotify_event)) + event->len;
                }
            }
        }
    } else
#endif /* USE_INOTIFY */
    {
        if (-1 == stat(subusers_file, &st)) {
            perror("stat " subusers_file);
            fprintf(stderr, "Could not stat " subusers_file "\n");
            fuse_exit(fuse_get_context()->fuse);           
        } else {
            if (last_mtime < st.st_mtime) {
/*fprintf(stderr, "pid %d tid %d: stat says we should update\n", getpid(), (int)pthread_self());*/
                subidfs_read_data();
/*fprintf(stderr, "pid %d tid %d: finished updating from stat\n", getpid(), (int)pthread_self());*/
            }
        }
    }

/*fprintf(stderr, "pid %d tid %d: releasing esc_mutex\n", getpid(), (int)pthread_self());*/
    pthread_mutex_unlock(&esc_mutex);    
/*fprintf(stderr, "pid %d tid %d: released esc_mutex\n", getpid(), (int)pthread_self());*/
}

static int xmp_evaluate_mode(const char* path, int mask) {
    int res;
    mode_t mode;
    struct stat stbuf;
    struct fuse_context *fc = fuse_get_context();

    res = lstat(path, &stbuf);
    if (res == -1)
        return -errno;

    mode = stbuf.st_mode;

    if (fc->uid == stbuf.st_uid) {
        mode >>= 6;
    } else if (fc->gid == stbuf.st_gid) {
        /* this is really only checking the user's current group.
         * in the kernel there's a routine for walking a btree of the
         * user's supplemental groups.  something like that should
         * go here, really. */
        mode >>= 3;
    }

/* taken from <linux/fs.h> */
#define MAY_EXEC 1
#define MAY_WRITE 2
#define MAY_READ 4
#define MAY_APPEND 8
    if (((mode & mask & (MAY_READ|MAY_WRITE|MAY_EXEC)) == mask))
        return 0;
    
    return EACCES;
}

static int xmp_subid_check(const char *path, int mask)
{
    int res;
    struct child_parent *subuser_data;
    struct stat stbuf;
    struct fuse_context *fc = fuse_get_context();

    res = lstat(path, &stbuf);
    if (res == -1)
        return -errno;

    update_etc_subusers_contents();
    subuser_data = etc_subusers_contents;
    if (subuser_data == NULL) {
        return -ENXIO; /* FIXME return something (more) meaningful */
    }

    if (is_strict_descendant_cached(subuser_data, fc->uid, stbuf.st_uid)) {
        /* grant access because the file belongs to a subuser of the caller */
        return 0;
    }

    /* res = access(path, mask); */
    /* instead of calling access(), we have to manually calculate 
     * the permission on the file according to fc->{u,g}id.  The
     * following is adapted from generic_permission() in 2.6.18. */
    res = xmp_evaluate_mode(path, mask);

    return -res;
}

static int xmp_ancestor_check(const char *path)
{
    int res;
    struct child_parent* subuser_data;
    struct stat stbuf;
    struct fuse_context *fc = fuse_get_context();

    res = lstat(path, &stbuf);
    if (res == -1)
        return -errno;

    update_etc_subusers_contents();
    subuser_data = etc_subusers_contents;
    if (subuser_data == NULL) {
        return -ENXIO; /* FIXME return something (more) meaningful */
    }

    if (is_descendant_cached(subuser_data, fc->uid, stbuf.st_uid)) {
        /* grant access because the file belongs to a subuser of the caller */
        return 0;
    }

    return -EACCES;
}

static int xmp_getattr(const char *path, struct stat *stbuf)
{
    int res;

    res = lstat(path, stbuf);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_access(const char *path, int mask)
{
    int res;

    res = xmp_subid_check(path, mask);
    if (!res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    res = access(path, mask);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_readlink(const char *path, char *buf, size_t size)
{
    int res;

    res = xmp_subid_check(path, MAY_READ);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = readlink(path, buf, size - 1);
    if (res == -1)
        return -errno;

    buf[res] = '\0';
    return 0;
}


static int xmp_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi)
{
    DIR *dp;
    struct dirent *de;
    int res;

    (void) offset;
    (void) fi;

    res = xmp_subid_check(path, MAY_READ|MAY_EXEC);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    dp = opendir(path);
    if (dp == NULL)
        return -errno;

    while ((de = readdir(dp)) != NULL) {
        struct stat st;
        memset(&st, 0, sizeof(st));
        st.st_ino = de->d_ino;
        st.st_mode = de->d_type << 12;
        if (filler(buf, de->d_name, &st, 0))
            break;
    }

    closedir(dp);
    return 0;
}

static int xmp_mknod(const char *path, mode_t mode, dev_t rdev)
{
    char* paths_dir = NULL;
    int res;
    struct fuse_context *fc = fuse_get_context();
    /* printf("Got fuse context.  uid=%d, gid=%d\n", fc->uid, fc->gid); */

    paths_dir = strdup(path);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    /* On Linux this could just be 'mknod(path, mode, rdev)' but this
       is more portable */
    if (S_ISREG(mode)) {
        res = open(path, O_CREAT | O_EXCL | O_WRONLY, mode);
        if (res >= 0)
            res = close(res);
    } else if (S_ISFIFO(mode)) {
        res = mkfifo(path, mode);
    } else {
        res = mknod(path, mode, rdev);
    }

    if (res == -1)
        return -errno;

    /* we're going to run the fs as root, and let everybody use it.  if
       we don't chown(2) the file, it would be owned by root.root (or
       whoever the fs is running as).  */
    res = chown(path, fc->uid, fc->gid);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_mkdir(const char *path, mode_t mode)
{
    char* paths_dir = NULL;
    int res;
    struct fuse_context *fc = fuse_get_context();

    paths_dir = strdup(path);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = mkdir(path, mode);
    if (res == -1)
        return -errno;

    /* we're going to run the fs as root, and let everybody use it.  if
       we don't chown(2) the file, it would be owned by root.root (or
       whoever the fs is running as).  */
    res = chown(path, fc->uid, fc->gid);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_unlink(const char *path)
{
    char* paths_dir = NULL;
    int res;

    paths_dir = strdup(path);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = unlink(path);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_rmdir(const char *path)
{
    char* paths_dir = NULL;
    int res;

    paths_dir = strdup(path);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = rmdir(path);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_symlink(const char *from, const char *to)
{
    char* paths_dir = NULL;
    int res;
    struct fuse_context *fc = fuse_get_context();

    paths_dir = strdup(to);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = symlink(from, to);
    if (res == -1)
        return -errno;

    /* we're going to run the fs as root, and let everybody use it.  if
       we don't chown(2) the file, it would be owned by root.root (or
       whoever the fs is running as).  */
    res = chown(to, fc->uid, fc->gid);
    if (res == -1)
        return -errno;

    return 0;
}

/* FIXME there's lots of corner cases here, check out vfs_rename */
static int xmp_rename(const char *from, const char *to)
{
    char* paths_dir = NULL;
    int res;

    /* check that we may_delete from the from directory */
    paths_dir = strdup(from);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    /* check that we may_create in the to directory */
    paths_dir = strdup(to);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    res = rename(from, to);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_link(const char *from, const char *to)
{
    char* paths_dir = NULL;
    int res;

    paths_dir = strdup(to);
    *(strrchr(paths_dir, '/')) = '\0';
#ifdef DEBUG
    printf("checking %s for MAY_WRITE|MAY_EXEC\n", paths_dir);
#endif
    res = xmp_subid_check(paths_dir, MAY_WRITE|MAY_EXEC);
    free(paths_dir);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    res = link(from, to);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_chmod(const char *path, mode_t mode)
{
    int res;

#ifdef DEBUG
    printf("checking whether %s is owned by the caller or an ancestor thereof\n", path);
#endif
    res = xmp_ancestor_check(path);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    res = chmod(path, mode);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_chown(const char *path, uid_t uid, gid_t gid)
{
    int res;

    /* check that path is owned by the caller or some ancestor thereof */
#ifdef DEBUG
    printf("checking whether %s is owned by the caller or an ancestor thereof\n", path);
#endif
    res = xmp_ancestor_check(path);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif

    /* FIXME:  check also that fc->uid is an ancestor of uid */
#ifdef DEBUG
    printf("/* FIXME:  check also that fc->uid is an ancestor of uid */\n");
#endif

    res = lchown(path, uid, gid);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_truncate(const char *path, off_t size)
{
    int res;

    res = xmp_subid_check(path, MAY_WRITE);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = truncate(path, size);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_utime(const char *path, struct utimbuf *buf)
{
    int res;

    res = xmp_subid_check(path, MAY_WRITE);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = utime(path, buf);
    if (res == -1)
        return -errno;

    return 0;
}


static int xmp_open(const char *path, struct fuse_file_info *fi)
{
    int res;

    res = xmp_subid_check(path, MAY_READ);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    res = open(path, fi->flags);
    if (res == -1)
        return -errno;

    close(res);
    return 0;
}

static int xmp_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
    int fd;
    int res;

    res = xmp_subid_check(path, MAY_READ);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    (void) fi;
    fd = open(path, O_RDONLY);
    if (fd == -1)
        return -errno;

    res = pread(fd, buf, size, offset);
    if (res == -1)
        res = -errno;

    close(fd);
    return res;
}

static int xmp_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    int fd;
    int res;

    res = xmp_subid_check(path, MAY_WRITE);
    if (res) {
        return res;
    }
#ifdef DEBUG
    printf("subid check succeeded in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); fflush(stdout);
#endif
    
    (void) fi;
    fd = open(path, O_WRONLY);
    if (fd == -1)
        return -errno;

    res = pwrite(fd, buf, size, offset);
    if (res == -1)
        res = -errno;

    close(fd);
    return res;
}

static int xmp_statfs(const char *path, struct statvfs *stbuf)
{
    int res;

    res = statvfs(path, stbuf);
    if (res == -1)
        return -errno;

    return 0;
}

static int xmp_release(const char *path, struct fuse_file_info *fi)
{
    /* Just a stub.  This method is optional and can safely be left
       unimplemented */

    (void) path;
    (void) fi;
    return 0;
}

static int xmp_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
    /* Just a stub.  This method is optional and can safely be left
       unimplemented */

    (void) path;
    (void) isdatasync;
    (void) fi;
    return 0;
}

#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int xmp_setxattr(const char *path, const char *name, const char *value,
                        size_t size, int flags)
{
    int res = lsetxattr(path, name, value, size, flags);
    if (res == -1)
        return -errno;
    return 0;
}

static int xmp_getxattr(const char *path, const char *name, char *value,
                    size_t size)
{
    int res = lgetxattr(path, name, value, size);
    if (res == -1)
        return -errno;
    return res;
}

static int xmp_listxattr(const char *path, char *list, size_t size)
{
    int res = llistxattr(path, list, size);
    if (res == -1)
        return -errno;
    return res;
}

static int xmp_removexattr(const char *path, const char *name)
{
    int res = lremovexattr(path, name);
    if (res == -1)
        return -errno;
    return 0;
}
#endif /* HAVE_SETXATTR */

static struct child_parent* new_cp() {
    struct child_parent* cp;
    cp = (struct child_parent*)malloc(sizeof(struct child_parent));
    if (!cp) {
        perror("Could not allocate memory\n");
        exit(1);
    }
    cp->child  = -1;
    cp->parent = -1;
    cp->next   = NULL;

    return cp;
}

void free_cp_list(struct child_parent* the_list) {
    struct child_parent* cp = the_list;
    struct child_parent* cp_old = NULL;

    while (NULL != cp) {
        cp_old = cp;
        cp = cp->next;
        free(cp_old);
    }
}

#define BUFFER_SIZE 1024
/* Entering this function requires locking &esc_mutex */
void subidfs_read_data() {
    struct child_parent *head;
    struct child_parent *walker;
    char buffer[BUFFER_SIZE] = {0};
    FILE* sufile = NULL;
    int i;
    int j;
    /* way, way overkill, but boo to str{tok,sep}. */
    char parent_name[BUFFER_SIZE] = {0};
    char child_name[BUFFER_SIZE] = {0};

    /* for getpwnam_r */
    struct passwd pd;
    struct passwd* pwdptr = &pd;
    struct passwd* temp_pwd_ptr;
    char pwdbuffer[BUFFER_SIZE];
    int  pwdlinelen = sizeof(pwdbuffer);
    uid_t parent_uid;

    head = new_cp();
    head->child  = -2;
    head->parent = -2;
    head->next   = NULL;

    walker = head;

    sufile = fopen(subusers_file, "r");
    if (!sufile) {
        perror("couldn't open subuser map file");
        fuse_exit(fuse_get_context()->fuse);
    }

    /* slurp in each line, allocating memory for each */
    while (fgets(buffer,BUFFER_SIZE,sufile)) {
        /* kill the newline that fgets leaves */
        if (buffer[strlen(buffer) - 1] != '\n') {
            fprintf(stderr, "last character of buffer was not a newline -- maybe BUFFER_SIZE isn't long enough?\n");
            fuse_exit(fuse_get_context()->fuse);
        }
        buffer[strlen(buffer) - 1] = '\0';

        i = 0;
        while (buffer[i] != ':') {
            parent_name[i] = buffer[i];
            ++i;
        }
        parent_name[i] = '\0';
        ++i;

        if (0 != (getpwnam_r(parent_name,pwdptr,pwdbuffer,pwdlinelen,&temp_pwd_ptr))) {
            perror("getpwnam_r() error");
            fuse_exit(fuse_get_context()->fuse);
        }
        parent_uid = pd.pw_uid;

        while (buffer[i] != '\0') {
            j = 0;
            while ((buffer[i] != ',') && (buffer[i] != '\0')) {
                child_name[j] = buffer[i];
                ++i;
                ++j;
            }
            child_name[j] = '\0';
            if (buffer[i] == '\0') {
                /* undo the next incr, so that the enclosing while 
                   doesn't walk off the end of the buffer */
                --i;
            }
            ++i;

            if (0 != (getpwnam_r(child_name,pwdptr,pwdbuffer,pwdlinelen,&temp_pwd_ptr))) {
                perror("getpwnam_r() error");
                fuse_exit(fuse_get_context()->fuse);
            }

            walker->next = new_cp();
            walker->next->child  = pd.pw_uid;
            walker->next->parent = parent_uid;
            walker = walker->next;
        }
    }
    fclose(sufile);

    if (NULL != etc_subusers_contents) {
        free_cp_list(etc_subusers_contents);
    }

    etc_subusers_contents = head;
    return;
}

static void* xmp_init(void)
{
    return (void*) NULL;
}

static struct fuse_operations xmp_oper = {
    .getattr	= xmp_getattr,
    .access	= xmp_access,
    .readlink	= xmp_readlink,
    .readdir	= xmp_readdir,
    .mknod	= xmp_mknod,
    .mkdir	= xmp_mkdir,
    .symlink	= xmp_symlink,
    .unlink	= xmp_unlink,
    .rmdir	= xmp_rmdir,
    .rename	= xmp_rename,
    .link	= xmp_link,
    .chmod	= xmp_chmod,
    .chown	= xmp_chown,
    .truncate	= xmp_truncate,
    .utime	= xmp_utime,
    .open	= xmp_open,
    .read	= xmp_read,
    .write	= xmp_write,
    .statfs	= xmp_statfs,
    .release	= xmp_release,
    .fsync	= xmp_fsync,
    .init   = xmp_init,
#ifdef HAVE_SETXATTR
    .setxattr	= xmp_setxattr,
    .getxattr	= xmp_getxattr,
    .listxattr	= xmp_listxattr,
    .removexattr= xmp_removexattr,
#endif
};

int main(int argc, char *argv[])
{
    umask(0);

    using_inotify = 0;
#ifdef USE_INOTIFY
    inotify_fd.fd = inotify_init();
    if (inotify_fd.fd < 0) {
        fprintf(stderr,"Could not initialize inotify\n");
    } else {
        wd = inotify_add_watch(inotify_fd.fd, subusers_file, IN_MODIFY );

        if (wd < 0) {
            fprintf(stderr, "Could not add watch on " subusers_file "\n");
        } else {
            inotify_fd.events = POLLIN;
            using_inotify = 1;
        }
    }
    if (!using_inotify) fprintf(stderr, "falling back to checking mtime\n");
    if (!using_inotify)
#endif /* USE_INOTIFY */
        if (-1 == stat(subusers_file, &st)) {
            perror("stat " subusers_file);
            fprintf(stderr,"Couldn't stat " subusers_file ", exiting\n");
            exit(1);
        }

    last_mtime = st.st_mtime;
    subidfs_read_data();

    return fuse_main(argc, argv, &xmp_oper);
}
