/* $Id: subuserdel.c 84 2007-01-26 06:38:21Z psnowber $
 * vim: set tw=80 ts=4 sw=4 cindent : */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#include "subid.h"
#include "log.h"
#include "util_uid.h"

void usage(int argc, char* argv[]) {
	printf("usage: %s    <user name>\n", argv[0]);
	printf("       %s -q <user name>\n", argv[0]);
	exit(1);
}

void usage_really(int argc, char* argv[]) {
	printf("Are you sure?  This utility deletes all the files belonging to <user name>.\n");
	printf("If you're really, really sure, the actual usage is as follows:\n");
	printf("usage: %s -r    <user name>\n", argv[0]);
	printf("       %s -r -q <user name>\n", argv[0]);
	exit(1);
}

int use_fully_qualified = 0;

int main(int argc, char* argv[]) {
	struct passwd* delete_pw = NULL;
	pid_t childpid;
	int childstatus;
	char* name_to_delete = NULL;
	char* calling_name = NULL;

	int really = 0;
	int c;
	int _optind = -1;

    while ((c = getopt(argc, argv, "qr")) != -1) {
        switch(c) {
        case 'q':
            use_fully_qualified = 1;
            break;
        case 'r':
            really = 1;
            break;
        default:
            usage(argc, argv); /* calls exit(1) */
        }
    }
    _optind = optind; /* easier to watch this with gdb */

    if (_optind >= argc) {
        /* the user didn't specify a username */
        usage(argc, argv); /* calls exit(1) */
    }
	if (!really) {
		usage_really(argc, argv); /* calls exit(1) */
	}	

	/* look up the username of the calling process with getuid() */
	{
		calling_name = getpwuid(getuid())->pw_name;
	}

	if (use_fully_qualified) {
		name_to_delete = (char*)malloc(sizeof(char)*
			(strlen(calling_name) + 1 + strlen(argv[_optind]) + 1));
		if (!name_to_delete) {
			fprintf(stderr, "couldn't allocate memory!");
			exit(1);
		}
		memset(name_to_delete, 0, strlen(calling_name) + 1 + strlen(argv[_optind]) + 1);
		strcat(name_to_delete, calling_name);
		strcat(name_to_delete, USERNAME_SEPARATOR);
		strcat(name_to_delete, argv[_optind]);
	} else {
		name_to_delete = argv[_optind];
	}
	
	{
		errno = 0; /* says the getpwnam man page */
		delete_pw = getpwnam(name_to_delete);
		if (!delete_pw) {
			/* careful -- we want the entry to not exist */
			switch (errno) {
				case ENOMEM:
				case EIO:
				case EINTR:
				case EMFILE:
				case ENFILE:
				case ERANGE:
					perror("main(): there was an error looking up that passwd entry: ");
					exit(127);
				default:
					fprintf(stderr,"there is no user with that username\n");
					exit(1);
			}
		} else { /* getpwnam returned a valid struct passwd* */
		}

	}

	/* must check that the user to be deleted is a subuser of the caller */
	if (!strcmp(calling_name, name_to_delete)) {
		fprintf(stderr, "You can't use this tool to delete yourself\n");
		exit(1);
	}
	if (!is_strict_descendant(getpwnam(calling_name)->pw_uid, 
		                      getpwnam(name_to_delete)->pw_uid)) {
		fprintf(stderr, "permission denied: that child is not one of your descendants\n");
		exit(1);
	}
	
	/* after here, we start actually affecting system files, so we need to be
	 * able to back out after any given step; enter the ugly goto section */

	if (remove_relation(getpwnam(name_to_delete)->pw_uid)) {
		fprintf(stderr, "couldn't change /etc/subusers\n");
		exit(1);
	}

	childpid = fork();
	if (-1 == childpid) {
		/* error */
		perror("could not fork child process");
		goto fork_failure;
	} else if (0 == childpid) {
		/* child */

		/* append an entry to /etc/passwd */
		/* create the home directory */
		execl("/usr/sbin/userdel","/usr/sbin/userdel","-r",name_to_delete,(char*)NULL);
		perror("execl");
		exit(1);
	}

	/* parent */
	waitpid(childpid, &childstatus, 0);
	if (!(WIFEXITED(childstatus) && (0 == WEXITSTATUS(childstatus)))) {
		/* the child failed for some reason */
		fprintf(stderr, "/usr/sbin/useradd failed\n");
		goto fork_failure;
	}

	/* the child exited correctly */

	goto success;
	
fork_failure:
	return 1;

success:
	log_fmt("user %s deleted sub-user %s", calling_name, name_to_delete);
	return 0;
}
