/* $Id: util.c 73 2007-01-25 22:42:27Z psnowber $
 * vim: set tw=80 ts=4 sw=4 : */

#include <stdio.h>
#include <stdlib.h>
#define __USE_GNU /* to get strndup */
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#include <stdarg.h>

#include "subid.h"
#include "util.h"

#define debug_level 1
void debug(int lev, char* msg) {
	if (lev >= debug_level)
		printf(msg);
}

int log_fmt(char* fmt, ...) {
	FILE *logfile = NULL;
	char *ts = NULL;
	time_t t;

	va_list ap;
	va_start(ap,fmt);

	/* put the current time into t */
	time(&t);
	/* get the time as a string */
	ts = ctime(&t);
	if (!ts) {
		fprintf(stderr,"can't find out the time\n");
		return 0;
	}
	/* strip the newline */
	*(index(ts,'\n')) = '\0';
	
	logfile = fopen(log_file_name,"a+");
	if (!logfile) {
		fprintf(stderr,"could not open the log file\n");
		return 0;
	}

	fprintf(logfile, "%s: ", ts);
	vfprintf(logfile, fmt, ap);
	fprintf(logfile, "\n");
	fclose(logfile);

	return 1;
}

static int check_child(char**, const char*, const char*, int, int);

#ifdef TESTING_UTIL
#define subusers_file "/tmp/subusers"
int main(int argc, char** argv) {
	char** lines = read_data();

	if (argc < 3) {
		printf("usage: %s parent descendant\n",argv[0]);
		exit(1);
	}

	printf("direct child: ");
	if (is_child(lines, argv[1], argv[2]))
		printf("yes\n");
	else
		printf("no\n");

	printf("indirect child: ");
	if (is_descendant(lines, argv[1], argv[2]))
		printf("yes\n");
	else
		printf("no\n");

	return 0;
}
#endif

int is_strict_descendant(char** lines, const char* p, const char* c) {
	return check_child(lines, p, c, 1, 0);
}

int is_strict_child(char** lines, const char* p, const char* c) {
	return check_child(lines, p, c, 0, 0);
}

int is_descendant(char** lines, const char* p, const char* c) {
	/* users are their own children */
	if (!strcmp(p,c)) 
		return 1;

	return check_child(lines, p, c, 1, 0);
}

int is_child(char** lines, const char* p, const char* c) {
	/* users are their own children */
	if (!strcmp(p,c)) 
		return 1;

	return check_child(lines, p, c, 0, 0);
}

char * find_parent(char** lines, const char* p, const char* c) {
	int i=0;
	char* cpp = NULL;
	char* cp = NULL;
	char* child = NULL;
	char* tmp = NULL;

	while ((cpp = lines[i++])) {
		/* if the line does not begin with p, followed by : */
		if (strncmp(p,cpp,strcspn(cpp,":")))
			continue;

		/* else, we're on p's line, so his children must be here */
		cp = cpp + (strcspn(cpp,":") + 1);
		while (cp < cpp + strlen(cpp)) {
			if (! strncmp(cp,c,strcspn(cp,",")))
				/* we've found c */
				return strndup(cpp, strcspn(cpp,":"));
			
			/* that one wasn't c, so */
			/* advance cp past the next ',' */
			cp = cp + (strcspn(cp,",") + 1);
		}

		if (1) {
			/* none of p's direct children were c, but check c's children */
			cp = cpp + (strcspn(cpp,":") + 1);
			while (cp < cpp + strlen(cpp)) {
				child = strndup(cp,strcspn(cp,","));
				if (!child) {
					perror("could not allocate memory");
					exit(1);
				}

				if ((tmp = find_parent(lines,child,c))) {
					free(child);
					return tmp;
				}
	 
				free(child);

				/* advance cp past the next ',' */
				cp = cp + (strcspn(cp,",") + 1);
			}
		}

		/* if we return 0 here, then we're disallowing having multiple lines per
		 * parent user.  to allow this, remove this return 0 */
		return NULL;
	}

	return NULL;
}

static int check_child(char** lines, const char* p, const char* c, int descend, int depth) {
	int i=0;
	char* cpp = NULL;
	char* cp = NULL;
	char* child = NULL;

	if (depth >= MAX_DEPTH) {
		/* cut off at some reasonable amount, to not break on ownership loops */
		/* (which shouldn't be possible, but here we are) */
		return 0;
	}

	while ((cpp = lines[i++])) {
		/* if the line does not begin with p, followed by : */
		if (strncmp(p,cpp,strcspn(cpp,":")))
			continue;

		/* else, we're on p's line, so his children must be here */
		cp = cpp + (strcspn(cpp,":") + 1);
		while (cp < cpp + strlen(cpp)) {
			if (! strncmp(cp,c,strcspn(cp,",")))
				/* we've found c */
				return 1;
			
			/* that one wasn't c, so */
			/* advance cp past the next ',' */
			cp = cp + (strcspn(cp,",") + 1);
		}

		if (descend) {
			/* none of p's direct children were c, but check c's children */
			cp = cpp + (strcspn(cpp,":") + 1);
			while (cp < cpp + strlen(cpp)) {
				child = strndup(cp,strcspn(cp,","));
				if (!child) {
					perror("could not allocate memory");
					exit(1);
				}

				if (check_child(lines,child,c,1, depth+1)) {
					free(child);
					return 1;
				}
	 
				free(child);

				/* advance cp past the next ',' */
				cp = cp + (strcspn(cp,",") + 1);
			}
		}

		/* if we return 0 here, then we're disallowing having multiple lines per
		 * parent user.  to allow this, remove this return 0 */
		return 0;
	}

	return 0;
}

void free_lines(char** lines) {
	int i = 0;
	char* cp = NULL;

	while ((cp = lines[i++])) {
		free(cp);
	}
	free(lines);
}

int write_data(char** lines) {
	FILE* sufile = NULL;
	char* cp = NULL;
	int i = 0;

	sufile = fopen(subusers_file, "w+");
	if (!sufile) {
		perror("couldn't open subuser map file for writing");
		return 0;
	}

	while ((cp = lines[i++])) {
		fprintf(sufile,"%s\n",cp);
	}

	fclose(sufile);

	return 1;
}

#define BUFFER_SIZE 1024
char** read_data() {
	char **lines = NULL;
	char buffer[BUFFER_SIZE] = {0};
	FILE* sufile = NULL;
	int numlines = 0;
	int readed;
	int i;

	sufile = fopen(subusers_file, "r");
	if (!sufile) {
		perror("couldn't open subuser map file");
		return NULL;
	}

	/* find out many lines are in the file */
	while (0 != (readed = fread(buffer,1,BUFFER_SIZE,sufile))) {
		if (ferror(sufile)) {
			fprintf(stderr,"error while reading the subuser map file\n");
			return NULL;
		}
		for (i=0;i<readed;i++) {
			if (buffer[i] == '\n')
				numlines++;
		}
		if (feof(sufile)) {
			break;
		}
	}

	/* allocate enough char*'s to get every line, plus a sentinal NULL */
	lines = (char **)malloc((numlines+1) * sizeof(char*));
	if (!lines) {
		perror("can't allocate memory");
		return NULL;
	}
	
	/* go back to read in all the lines */
	clearerr(sufile); rewind(sufile);

	/* slurp in each line, allocating memory for each */
	for (i=0;i<numlines;i++) {
		fgets(buffer,BUFFER_SIZE,sufile);
		/* kill the newline that fgets leaves */
		buffer[strlen(buffer) - 1] = '\0';
		
		lines[i] = strdup(buffer);
		if (!lines[i]) {
			perror("can't allocate memory");
			return NULL;
		}
	}
	/* sentinal NULL */
	lines[numlines] = NULL;

	fclose(sufile);
	return lines;
}
